/*  Sony RM-X2S plugin for xmms by Dirk Jagdmann <doj@cubic.org>
 *                                 http://llg.cubic.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.  */

/* $Header: /home/cvs/sonyjoy/calibrate.c,v 1.9 2006/05/05 20:10:33 doj Exp $ */

#include <fcntl.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>

#include "sony.h"

/** number of steps */
#define NUM_POS 3

/** string to show in the dialog for each step of the calibration procedure */
static const char *pos_name[] =
{
  "Hold OFF Button on Sonystick, then click OK still pressing OFF",
  "Hold MUTE Button on Sonystick, then click OK still pressing MUTE",
  "Leave Sonystick in Home/Default Position, then click OK",
  "nil"
};
/** the minimum values for each step */
static int cmin[NUM_POS];
/** the maximum values for each step */
static int cmax[NUM_POS];
/** the filehandle for the joystick port */
static int fd=-1;
/** the linux joystick driver correction structure */
static struct js_corr corr;
/** number of buttons */
static int buttons;
/** number of axis */
static int axis;

/** window widget */
static GtkWidget *window=0;
/** status bar widget */
static GtkWidget *status=0;
/** the current step */
static int turn=0;

/** @return the current time in miliseconds */
static int sonyrm_get_time()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

/** wait for an event on the joystick port. aborts after 1/10 seconds */
static void sonyrm_wait_for_event()
{
  struct timeval tv;
  fd_set set;

  if(fd < 0)
    return;

  tv.tv_sec = 0;
  tv.tv_usec = 100000;

  FD_ZERO(&set);
  FD_SET(fd, &set);

  if (select(fd+1, &set, NULL, NULL, &tv))
    {
      if (FD_ISSET(fd, &set))
	{
	  struct js_event ev;
	  if (read(fd, &ev, sizeof(struct js_event)) == sizeof(struct js_event))
	    switch (ev.type & ~JS_EVENT_INIT)
	      {
	      case JS_EVENT_AXIS:
		axis = ev.value; break;
	      case JS_EVENT_BUTTON:
		buttons = (buttons & ~(1 << ev.number)) | (ev.value << ev.number);
	      }
	}
    }
  else
    {
      buttons &= ~(1 << 31);
    }
}

/** display a text on the status bar
    @param text the string to display
*/
static void addstatus(const gchar *text)
{
  gtk_statusbar_push(GTK_STATUSBAR(status),
		     gtk_statusbar_get_context_id(GTK_STATUSBAR(status), text),
		     text);
}

/** The OK button was pressed. Now we have to calibrate the joystick
    for the current step
    @param w ignored
    @param data pointer to sonyrm configuration data
*/
static void sonyrm_button_ok(GtkWidget * w, gpointer data)
{
  int t;
  sonyrm_config *cfg=(sonyrm_config *)data;

  /* clear events in the read cue */
  t = sonyrm_get_time();
  while (sonyrm_get_time() < t + 50)
    sonyrm_wait_for_event(fd);

  cmin[turn] = axis;
  cmax[turn] = axis;

  /* measure for 2sec */
  t = sonyrm_get_time();
  while (sonyrm_get_time() < t + 2000)
    {
      sonyrm_wait_for_event(fd);
      if (axis < cmin[turn]) cmin[turn] = axis;
      if (axis > cmax[turn]) cmax[turn] = axis;
    }

  turn++;
  addstatus(_(pos_name[turn]));
  putchar(7); fflush(stdout);	/* beep */

  if(turn >= NUM_POS)
    {
      /* we are finished. calculate correction and copy the config */
      int i;
      double a, b;

      a=32767.0 / (cmin[1] - cmax[0]);
      b=32767.0 / (cmin[2] - cmax[1]);

      corr.coef[0] = cmin[1];
      corr.coef[1] = cmax[1];
      corr.coef[2] = (a * 16384.0);
      corr.coef[3] = (b * 16384.0);

      corr.type = JS_CORR_BROKEN;

      if (ioctl(fd, JSIOCSCORR, &corr))
	{
	  perror("jscal: error setting correction");
	  return;
	}

      close(fd);
      fd=-1;

#ifdef debug
      printf("sonyrm_button_ok(): type: %i\n", corr.type);
      printf("sonyrm_button_ok(): prec: %i\n", corr.prec);
      for(i=0; i<JOYSTICK_MAX_AXIS; i++)
	printf("sonyrm_button_ok(): coef[%i]: %ul\n", i, corr.coef[i]);
#endif

      cfg->correction.type=corr.type;
      cfg->correction.prec=corr.prec;
      for(i=0; i<JOYSTICK_MAX_AXIS; i++)
	cfg->correction.coef[i]=corr.coef[i];

      gtk_widget_destroy(window);
      sonyrm_addstatus(_("Now set the precision"));
    }

}

/** the cancel button was pressed. we abort calibration.
    @param w ignored
    @param data ignored
*/
static void sonyrm_button_cancel(GtkWidget * w, gpointer data)
{
  gtk_widget_destroy(window);
  if(fd>=0)
    if(close(fd) < 0)
      perror("error while closing device");
  fd=-1;
}

/** initialize the dialog and the joystick port
    @param w ignored
    @param data pointer to sonyrm configuration data
*/
void sonyrm_calibrate_init(GtkWidget * w, gpointer data)
{
  GtkWidget *button, *box, *label;
  int t;
  int amax, amin;
  sonyrm_config *cfg=(sonyrm_config *)data;

  turn=0;

  /* now read the initial joystick values and calculate precision */
  if ((fd = open(cfg->device, O_RDONLY)) < 0)
    {
      perror("jscal: can't open joystick device");
      return;
    }

  corr.type = JS_CORR_NONE;
  corr.prec = 0;

  if (ioctl(fd, JSIOCSCORR, &corr))
    {
      perror("jscal: error setting correction");
      close(fd);
      return;
    }

  /* Calibrating precision: wait and don't touch the joystick. */

  /* clear events in read cue */
  t = sonyrm_get_time();
  while (sonyrm_get_time() < t+50)
    sonyrm_wait_for_event(fd);

  amin = amax = axis;

  /* measure for 2sec */
  t = sonyrm_get_time();
  while (sonyrm_get_time() < t+2000)
    {
      sonyrm_wait_for_event(fd);
      if (amin > axis) amin = axis;
      if (amax < axis) amax = axis;
    }
  corr.prec = amax - amin;

  /* open the window */
  window=gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_window_set_title(GTK_WINDOW(window), _("Sonystick Calibration"));
  gtk_window_set_policy(GTK_WINDOW(window), FALSE, FALSE, TRUE);
  gtk_window_set_modal(GTK_WINDOW(window), TRUE);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
  gtk_widget_show(window);

  box=gtk_vbox_new(TRUE, 0);
  gtk_container_add(GTK_CONTAINER(window), box);
  gtk_widget_show(box);

  label=gtk_label_new(_("Note: Each measurement takes 2sec."));
  gtk_box_pack_start_defaults(GTK_BOX(box), label);
  gtk_widget_show(label);

  status=gtk_statusbar_new();
  gtk_box_pack_start_defaults(GTK_BOX(box), status);
  gtk_widget_show(status);
  gtk_widget_set_usize (status, 400, 20);

  button=gtk_button_new_with_label(_("OK"));
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_button_ok), cfg);

  button=gtk_button_new_with_label(_("Cancel"));
  gtk_box_pack_start_defaults(GTK_BOX(box), button);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(sonyrm_button_cancel), NULL);

  addstatus(_(pos_name[turn]));
}
